%% CLPQ support for Curry.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module('Clpqlib', []).

:- use_module('Core', [tonf/2, tohnf/2]).

:- multifile     user_tohnf/2, user_tonf/2.
:- discontiguous user_tohnf/2, user_tonf/2.

%% We are going to support clpq.
:- use_package(clpq).

%% Integer operations.
user_tohnf('Clpq_:+:'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B .=. F+G.

user_tohnf('Clpq_:-:'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B .=. F-G.

user_tohnf('Clpq_:*:'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B .=. F*G.

%% Comparison operators.

user_tohnf('Clpq_:<:'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	F .<. G,
	B = 'Prelude_Success'.

user_tohnf('Clpq_:>:'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	F .>. G,
	B = 'Prelude_Success'.

user_tohnf('Clpq_:>=:'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	F .>=. G,
	B = 'Prelude_Success'.

user_tohnf('Clpq_:<=:'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	F .=<. G,
	B = 'Prelude_Success'.
