%% Core module.
%%
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Julio Mariño Carballo
%%   * Jose María Rey Poza
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module('Core', [tonf/2, tohnf/2]).

%% Sloth's main curry engine.

:- use_module(library(lists),         [reverse/2] ).

:- multifile     user_tohnf/2, user_tonf/2, cons_meta/3.
:- discontiguous user_tohnf/2, user_tonf/2, cons_meta/3.

%% This operator is used by core and trans_lazy.
:- op(500, xfx, #).

%% To Head Normal Form translation.

%% Be carefull as we can have some infite terms here, as 
%% tohnf(main#R, R).  

%% Variables
tohnf(A, B) :- 
	var(A),
	!,
	A = B.

%% Numbers.
tohnf(A, B) :- 
	number(A),
	!,
	A = B.

%% Constructors.
tohnf(A, B) :-
	functor(A, Cons, _),
	cons_meta(Cons, _, _),
	!,
	A = B.

%% Lazy eval.
tohnf(_#A, B) :-
        \+ var(A),
 	!,
 	A = B.

%% A partial applied function is in head normal form already.
tohnf(pap(F, N, L), pap(F, N, L)) :-
	!.
 
%% Lazy application.

tohnf(apply_list(A,B)#C, C) :-
	A = pap(_,_,_),
	!,
        eval_apply_list(A, B, C).

tohnf(apply_list(A, B)#C, C) :-
	tohnf(A, A1),
	tohnf(apply_list(A1, B)#C, C).

tohnf(A, B) :-
	user_tohnf(A, B).

%%*******************************************
%% REMEMBER THAT apply_list IS NEVER IN HNF *
%%*******************************************

%% Dealing with High-Order.
%% We've got 3 cases. Just a normal app, 
eval_apply_list(pap(E, 0, A), [], R) :-
	!,
	eval_complete_app(pap(E, 0, A), R).
eval_apply_list(pap(E, 0, A), L, R) :-
	!,
	eval_complete_app(pap(E, 0, A), T),
	AppL = apply_list(T, L),
	tohnf(AppL#R, R).
eval_apply_list(pap(E, N, A), [], pap(E, N, A)).
eval_apply_list(pap(E, N, A), [B|C], D) :-
	N1 is N - 1,
	eval_apply_list(pap(E, N1, [B|A]), C, D).

%% This is the final step.
eval_complete_app(pap(A,0,B), C) :- 
        reverse(B, D),
        E =.. [A|D], 
        (
	    cons_meta(A, _, _)->
	    F = tohnf(E, C)
	;
	    F = tohnf(E#C, C)
	),
        call(F).

% tonf rules.
tonf(A, B) :-
	var(A),
	!,
	A = B.

tonf(A, B) :-
	number(A),
	!, 
	A = B.

% Lazy eval.       
% Lazy eval. 
tonf(_A#B, C) :-
	\+ var(B),
	!,
	tonf(B, C).

tonf(A#B, C) :-
        tohnf(A#B, D),
        tonf(D, C).

tonf(A, B) :-
	user_tonf(A, B).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

