-- Monadic IO support for Curry.
--
-- Copyright (C) 2005-2006, Babel Group
--
-- Written by: 
--   * Emilio Jesús Gallego Arias
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

module IO where

import engine IOlib

infixl 1 >>, >>=
-- This is our IO library

-- World is just a fake type.
data IO a = World a

(>>=) :: IO a -> (a -> IO b) -> IO b
(World d) >>= f = f d

(>>) :: IO a -> IO b -> IO b
a >> b = a >>= (\_ -> b)

return a = World a

-- Puts a string.
putStr :: [Char] -> IO ()
putStr external

putStrLn :: [Char] -> IO ()
putStrLn external

putInt :: Int -> IO ()
putInt external

getChar :: IO Char
getChar external

getLine = getChar >>= (\c ->
  if c=='\n' then return []
             else getLine >>= (\cs -> return (c:cs)))

print :: a -> IO ()
print e = putStr (show e)

printLn :: a -> IO ()
printLn e = putStrLn (show e)

show :: a -> [Char]
show external
