%% Sloth bootstrap code.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(gencore, [main/0]).

:- use_module(library(aggregates),    [findall/3]).
:- use_module(library(terms),         [atom_concat/2] ).

%% FIXME: Should use compiler modules.

% Type operators.
:- op(651, xfy, <@).  % Typevar aplication.
:- op(652, xfy, -->). % Normal apply operator.
:- op(653, xfy, =>).  % Type class constraint.

%% We have to bootstrap symbol table info for the types that we cannot
%% define in Curry proper.

main :- 
	create_core_exports.

module_header :-
	portray_clause((:- module('CoreTypes', [exports/1]))),
	portray_clause((:- discontiguous tohnf/2, tonf/2, cons_meta/3, exports/1)),
	portray_clause((:- op(651, xfy, <@))),  % Typevar aplication.
	portray_clause((:- op(652, xfy, -->))), % Normal apply operator.
	portray_clause((:- op(653, xfy, =>))).  % Type class constraint.


create_core_exports :-
	module_header,	
	%% Atomic types: Int, Char.
	export_type(tid('Prelude', 'Int')<@[],  []),
	export_type(tid('Prelude', 'Char')<@[], []),
	%% Null type.
        export_type(tid('Prelude', 'void')<@[],
	             [(qid('Prelude', '()'), tid('Prelude', 'void')<@[])]),
	%% Lists
	export_type(tid('Prelude', [])<@[A], 
	             [(qid('Prelude', []),  tid('Prelude', [])<@[_]),
		      (qid('Prelude', ':'), A --> tid('Prelude', [])<@[A] 
		                              --> tid('Prelude', [])<@[A])]),
        %% Tuples.
        export_type(tid('Prelude', tuple)<@[A, B], 
	             [(qid('Prelude', ','), A --> B --> tid('Prelude', tuple)<@[A,B]) ] ),
	%% FIXME: Should be a type alias, 
        export_type(tid('Prelude', 'String')<@[], []),
	!.
	
%% Exporting a type is now super easy :)
export_type(Name, ConsL) :-
	portray_clause(exports(type(Name, ConsL))),
	!.

%% We let the Curry compiler take care of translating constructors
%% when compiling the Prelude.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

