%% Translation for module's interfaces.
%%
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(trans_interfaces, [write_interfaces/1]).


:- use_module(symtab,              [get_object_property/3]).

%%
:- use_module(library(write),      [portray_clause/1] ).



write_interfaces(Module) :-
	findall(Imp, get_object_property(Module, imports, Imp), Imports),
	write_imports(Imports),
	nl,
	get_object_property(Module, exports, Exports),
	write_exports(Exports).

%% This is needed, or when a module have no exports CIAO will complain
%% about missing predicate exports.
write_exports([]) :- 
	portray_clause(exports(end)),
	!.

write_exports([(Name, Cons)|Exp]) :-
	export_cons(Cons, ConsExp),
 	portray_clause(exports(type(Name, ConsExp))),
	!,
 	write_exports(Exp).

write_exports([F|Exp]) :-
	get_object_property(F, type, Type),
	( 
	    get_object_property(F, assoc, Assoc),
	    get_object_property(F, priority, Prio) ->
	    OpAttributes = op(Assoc, Prio)
	;
	    OpAttributes = none
	),
	portray_clause(exports(id(F, Type, OpAttributes))),
	write_exports(Exp).

export_cons([], []).
export_cons([C|CN], [(C, CType)|CEN]) :-
	get_object_property(C, type, CType),
	export_cons(CN, CEN).

%%
write_imports(L) :-
	clause_write_imports(L, Imports),
	portray_clause((:- initialization(Imports))).

%%
clause_write_imports([], (!)).
clause_write_imports([M|Rest], (use_module(sloth(M)), Imports) ) :-
	clause_write_imports(Rest, Imports).
