%% Curry to prolog main translator.
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Julio Mariño Carballo
%%   * Jose María Rey
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(trans_lazy, [trans_module/1, trans_module_named/2], [assertions]).

:- use_module(deftree,       [defTree/3, call_defTree/2] ).
:- use_module(symtab,        [get_rules/2, get_object_property/3, 
	                      object_arity/2, get_module_functions/2] ).
:- use_module(trans_cons,    [translate_constructors/2] ).
:- use_module(trans_interfaces,    
                             [write_interfaces/1] ).
:- use_module(curryast,      [get_function/2] ).
:- use_module(const_expr,    [is_constant/1]).
:- use_module(cmdopts,       [cmd_options/2] ).

:- use_module(trans_tools,   [make_name/2] ).
:- use_module(trans_interfaces,    
	                     [write_interfaces/1] ).
:- use_module(trans_expr,    [trans_expr/2, parser2trad/2]).

%
:- use_module(library(lists),      [append/3, length/2, reverse/2, 
                                    list_concat/2] ).
:- use_module(library(aggregates), [findall/3] ).
:- use_module(library(write),      [write/1, write_term/2, portray_clause/1] ).
:- use_module(library(terms),      [atom_concat/2]).

%% FIXME: This is about to be killed.
:- include(ast_ops).
:- include(type_ops). %% Arggg.

%% subst_term(-Term, -Pos, -Exp, +NExp)
%% Pos is a list of positions.
subst_term(_Term,[],Subterm,Subterm) :-
	!.

subst_term(Term, Pos, Subterm, Term2) :-
	functor(Term, Functor, Arity),      
		%Cambia el termino en la posicion indicada
	functor(Term2, Functor, Arity),     % y lo sustituye por Subterm.
	subst_term_aux(Term, Term2, Pos, Subterm, Arity).

% FIXME: Document this.
subst_term_aux(_, _, _, _, 0) :-
	!.
subst_term_aux(Term, Term2, [N|Ps], Subterm, N) :-
	!,
	arg(N,Term,T),
	subst_term(T,Ps,Subterm,Term3),
	arg(N,Term2,Term3),
	N1 is N-1,
	subst_term_aux(Term,Term2,[N|Ps],Subterm,N1).

subst_term_aux(Term,Term2,[N|Ps],Subterm,N2) :-
	\+ ( N = N2 ),
	arg(N2,Term,T),
	arg(N2,Term2,T),
	N1 is N2-1,
	subst_term_aux(Term,Term2,[N|Ps],Subterm,N1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Aux predicates.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Create fresh vars. Must see if CIAO has a builtin pred for doing this.
%% We can use length(Res, N), it think.

new_vars(0, []) :- !.
new_vars(N, [_|Ts]) :-
	N1 is N-1,
	new_vars(N1, Ts).

%% This is like arg/3 but structurally recursive
term_at(Term, [], Term).
term_at(Term, [Pos|Poss], Subterm):-
	arg(Pos, Term, Term2),
	term_at(Term2, Poss, Subterm).

%% Generates a [[3],[2],[1]] list.
comienzos(0, []) :- !.
comienzos(N, [[N]|R]) :-
	N > 0,
	N1 is N-1,
	comienzos(N1, R).

%% Suffix a number or letter to the end of Pred.
letra_nivel(Pred, 0, Pred).
letra_nivel(Pred, Nivel, PredNivel):-
	name(Pred, PredList),
	Letra is 0'a + Nivel - 1,
	append(PredList, [Letra], PredList2),
	name(PredNivel, PredList2).

numero_nivel(Pred, Nivel, PredNivel) :-
	name(Pred, PredList),
	Letra is 0'0 + Nivel ,
	append(PredList, [Letra], PredList2),
	name(PredNivel, PredList2).
	
%% Get the name of the deftree translation for a predicate.
%% deftree_pred_name(-Pred, -Level, +NewPred)
deftree_pred_name(Pred, Level, NewPred) :-
	name(Pred, PredStr),
	(
	    Level = 0 ->
	    Sufijo = [0'1]
        ;
	    Letra is 0'a + Level - 1,
	    Sufijo = [0'1, Letra]
        ),
	append(PredStr, Sufijo, PredStr1),
	name(NewPred, PredStr1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% Part 1, tohnf translation.
trans_tohnf(F, Res) :-
	get_rules(F, Rt),
	(
	    Rt = [builtin(_)] ->
	    Res = []
	;
	    call_defTree(F, Tree),
	    tohnf_entry(F, Tree, 0, Res)
	).

%%% tohnf_entry(-Pred, -Deftree, -NumLetra, +Rule)
%%% Rule is a list of prolog clauses.
%%%
%%% This pred calls the deftree for a function, assuring
%%% that the branch arg is in hnf.
tohnf_entry(Pred, Deftree, NumLetra, Rule) :- 
	Head = user_tohnf(TermPred#Res, Res),  
	object_arity(Pred, Arity),
	new_vars(Arity, Vars),
	make_name(Pred, NPred),
	TermPred =.. [NPred|Vars],
	(
	    Deftree = branch(Pos, _)-_ ->
	    %% Get the argument that deftree is going to
	    %% switch on and tohnf it.
	    Rule = [(Head :- Body)],
	    term_at(TermPred, Pos, Var),
	    Body = (
		       tohnf(Var, Free1), 
		       TermResult
		   ),
            append(Vars, [Res], Vars2),
	    deftree_pred_name(NPred, NumLetra, PredDefTree),
	    TermResult_aux =.. [PredDefTree | Vars2],
	    subst_term(TermResult_aux, Pos, Free1, TermResult)
	;
	    Deftree = regla-_ ->
	    deftree_pred_name(NPred, NumLetra, PredDefTree),
	    append(Vars, [Res], Vars2),
	    BodyR =.. [PredDefTree | Vars2],
	    Rule  = [(Head :- BodyR)]
	;
	    Deftree = or(Deftrees)-_ ->
	    map_or_trees(Pred, Deftrees, 0, Rule) 
	;
	    Rule = [(Head :- fail)]
	).

%% map_or_trees(Pred,Arboles,Num_Rama,Reglas): Aplica tohnf_entry a
%% cada una de las alternativas de un arbol or y devuelve el resultado
%% en reglas (una lista de ellas).
map_or_trees(_, [], _, []).
map_or_trees(Pred, [T|Ts], N, Z) :-
	N1 is N+1,
	tohnf_entry(Pred, T, N, R),
	map_or_trees(Pred, Ts, N1, Rs),
	append(R, Rs, Z).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 2, deftree translation.

%% Can trans_deftree fail?
trans_deftree(F, Res):-
	get_rules(F, R),
	(
	    R = [builtin(_)] ->
	    Res = []
	;
	    (
		trans_deftree_real(F, Res) ->
		true
	    ;
		throw(error(translator, 'Cannot generate a definitional tree'(F)))
	    )
	).

%% This predicate translates the definitional tree of a predicate.
%% It has three cases: branch, or and rule.
trans_deftree_real(Pred, Reglas):-
	call_defTree(Pred, Deftree),
	Deftree = or(Arboles)-_,
	!,
	object_arity(Pred, Arity),
	new_vars(Arity, Vars),
	make_name(Pred, NPred),
	T_Acum =.. [NPred|Vars],
	map_trans_deftree_aux(Pred, 1, T_Acum, Arboles, 0, ReglasAux),
	list_concat(ReglasAux, Reglas).

trans_deftree_real(Pred, Reglas):-
	object_arity(Pred, Arity),
	get_rules(Pred, ReglasPred),
	comienzos(Arity, Comienzos),
	new_vars(Arity, Vars),
	make_name(Pred, NPred),
	TermAcum =.. [NPred|Vars],
	defTree(ReglasPred, Comienzos, DTree),
	!,
	(
	    DTree = branch(Pos,Arbol)-_Reglas ->
	    trans_deftree_aux(Pred, 1, TermAcum, (Pos,Arbol), 0, Reglas) 
	;
	    DTree = regla-[N] ->
	    trans_deftree_aux(Pred, 1, TermAcum, regla-[N], 0, Reglas)
	).  

%%% trans_deftree_aux(-Predicado, -Nivel, -Termino_acum, -Arbol, -NumArbol, +Reglas).
%%%
%%% The only thing this predicate does is to generate the number and the 
%%% name of predicates required for the tranlation, performing the actual 
%%% body translation by tr_rule.
trans_deftree_aux(_Predicado, _Nivel, _T_Acum, fallo-[], _, []) :- !.

trans_deftree_aux(Pred, Nivel, T_Acum, (or(Arboles)-_), _NumArbol, Reglas) :-
	!,
	map_trans_deftree_aux(Pred, Nivel, T_Acum,Arboles, 0, Reglas2),
	list_concat(Reglas2, Reglas).
  
trans_deftree_aux(Pred, Nivel, _T_Acum, regla-[Reg], NumArbol, [(Head :- Body)]) :-
	!,
	make_name(Pred, NPred),
	numero_nivel(NPred, Nivel, PredNumber),
	tr_rule(Reg, (Head1 :- Body)),
	Head1 =.. [_ | Args],
	letra_nivel(PredNumber, NumArbol, PredNb),
	Head =.. [PredNb | Args].

trans_deftree_aux(Pred, Nivel, T_Acum, branch(Pos, Ramas)-_, NumArbol, [(Head :- Body) | Rs]) :-
	make_name(Pred, NPred),
	N1 is Nivel + 1,
	numero_nivel(NPred,  Nivel,    PredN),
	letra_nivel( PredN,  NumArbol, PredNb),	
	numero_nivel(NPred,  N1,       PredN1),	
	letra_nivel( PredN1, NumArbol, PredN1b),
	
	term_at(T_Acum, Pos, T1),
	
        T_Acum =.. [_|T_AcumList],
	append(T_AcumList, [Free], HeadPars),
	
	Head =.. [PredNb | HeadPars], 

	subst_term(T_Acum, Pos, Free2, T_Acum2),
	T_Acum2 =.. [_|T_Acum2List],
	append(T_Acum2List, [Free], Head2Pars),
	letra_nivel(PredN1, NumArbol, PredN1b),
	TermPred1 =.. [PredN1b|Head2Pars],	
	Body = (
		   tohnf(T1, Free2),
		   TermPred1
	       ),
	N1 is Nivel + 1,
	trans_deftree_aux(Pred, N1, T_Acum, (Pos, Ramas), NumArbol, Rs),
	!.

trans_deftree_aux(Pred, Nivel, T_Acum, (Pos,[(Cons,Tree)|Ts]), NumArbol, Rs) :- !,
	(
	    is_constant(Cons) ->
	    ArityCons = 0 
	; 
	    object_arity(Cons, ArityCons)
	), 
	make_name(Cons, NCons),
	new_vars(ArityCons, Vars),
	TermCons =.. [NCons|Vars],
	subst_term(T_Acum, Pos, TermCons, T_Acum2),
	trans_deftree_aux(Pred, Nivel, T_Acum2, Tree, NumArbol, Rs1),
	trans_deftree_aux(Pred, Nivel, T_Acum, (Pos, Ts), NumArbol, Rs2),
	append(Rs1, Rs2, Rs3),
	Rs = Rs3.

%% Should output error or fail.
trans_deftree_aux(_,_,_,(_,[]),_,[]).

%% Only used for or- type deftrees.
map_trans_deftree_aux(_,_,_,[],_,[]).
map_trans_deftree_aux(Pred,Nivel,T_Acum,[Tree|Trees],NumArbol,Reglas):-
	NumArbol1 is NumArbol + 1,
	trans_deftree_aux(Pred, Nivel, T_Acum, Tree, NumArbol, ReglasTree),
	map_trans_deftree_aux(Pred, Nivel, T_Acum, Trees, NumArbol1, ReglasTrees),
	append([ReglasTree], ReglasTrees, Reglas).

%%%%%%%%%%%
% Rule translation.

% Caso 1, la regla empieza con un constructor.
add_arguments(Pred, VarList, NPred) :-
	Pred  =.. [P|ArgL],
	append(ArgL, VarList, NArgL),
	NPred =.. [P|NArgL].

%% We generate tohnf(pred(P1..Pn), 'Cons(Q1...Qn)') 
tr_rule(R, Rt) :-
	R = (Head:=Body),
	(
	    ( Body = qid(M, C)@_ )
	;
	    ( Body = qid(M, C) ) 
	),
	get_object_property(qid(M,C), constructs, _Type),
	!,
	trans_expr(Body, BodyTrad),
	parser2trad(Head, HeadTrad),
	add_arguments(HeadTrad, [BodyTrad], Ht),
	Rt = (Ht :- true).

%% Function translation. 
tr_rule(R,Rt) :-
	R = (Head := Body),
	get_function(Head, QPred),
	parser2trad(Head, HeadTrad), % Translates the pattern in the head.
	add_arguments(HeadTrad, [R0], HeadTrad2),
	trans_expr(Body, BodyTrad),
	BodyTrad2 = (
			tohnf(BodyTrad, R0)
		    ),
	% Este es el caso en el cual la cabecera de la función toma
	% menos argumentos que los necesarios -> debemos añadir más
	% variables. ¿Quizás debería ser tratado en otra parte?
	Rt_intermedio = (HeadTrad2 :- BodyTrad2),  
	(
	    get_rules(QPred, _Rules),
	    object_arity(QPred, TypeArity),
	    HeadTrad =.. [_|Hlist], 
	    length(Hlist, RuleArity),
	    TypeArity > RuleArity ->                    
	    Dif is TypeArity - RuleArity,
	    new_vars(Dif, Vars),
	    append(Vars, [R0], NVars),
	    add_arguments(HeadTrad, NVars, Head3),
            Body3 = (
			tohnf(apply_list(BodyTrad, Vars)#R0, R0)
		    ),
	    Rt = (Head3 :- Body3)
	;
	    Rt = Rt_intermedio
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Header of the module.

write_module_header(File, Module) :-
	(
	    get_rules(qid(Module, main), _) ->
	    ModuleExports = [exports/1, main/0]
	;
	    ModuleExports = [exports/1]
	),
	portray_clause((:- module(File, ModuleExports))),
	nl,
	static_header,
	nl,
	write_search_path,
	nl,
	write_core.

write_core :-
	cmd_options('-l', [LibDir]),
	atom_concat([LibDir, '/', 'Core'], CorePath),
	portray_clause((:- use_module(CorePath, [tonf/2, tohnf/2]))).

%%
write_search_path :-
	cmd_options('-l', [LibDir]),
	portray_clause(file_search_path(sloth, LibDir)),
	(
	    LibDir = '.' -> % Include . in the search path for normal files.
	    true
	;
	    portray_clause(file_search_path(sloth, '.'))
	).

%%
static_header :-
	portray_clause((:- use_module(library(system)))),
	portray_clause((:- use_module(library(lists)))),
	nl,
	portray_clause((:- dynamic       file_search_path/2)),
	portray_clause((:- multifile     cons_meta/3, user_tonf/2, user_tohnf/2, file_search_path/2)),
	portray_clause((:- discontiguous cons_meta/3, user_tonf/2, user_tohnf/2)),
	nl,
	%% For types export, umm this should done automatically??
	portray_clause((:- op(651, xfy, <@))),  % Typevar aplication.
	portray_clause((:- op(652, xfy, -->))), % Normal apply operator.
	portray_clause((:- op(653, xfy, =>))).  % Type class constraint.

%% Main translation function.
trans_module(Module) :-
	trans_module_named(Module, Module).

trans_module_named(File, Module) :-
	write_module_header(File, Module),
	!,
	write_interfaces(Module),
	!,
	translate_constructors(Module, ConsP), write_prolog(ConsP),
	!,
	trans_rules(Module), nl, 
	!,
	write_main_goal.

trans_rules(Module) :-
	get_module_functions(Module, Functions),
	trans_functions(Functions).
			   
trans_functions([]).
trans_functions([F|Fl]) :-
	trans_function(F, R),
	write_prolog(R),
	!,
	trans_functions(Fl).
	
trans_function(F, Res) :-
	trans_tohnf(F, FHnf),
	!,
	trans_deftree(F, FDefTree),
	append(FHnf, FDefTree, Res).

%% Temp:
write_prolog([]).
write_prolog([C|CL]) :-
	portray_clause(C), 
	nl,
	write_prolog(CL).

%% If we have a main function, link it with the prolog predicate main.
%% Fixme: Type checking doesn't belong here.
write_main_goal :-
	get_rules(qid(Module, main), _),
	!,
	get_object_property(qid(Module, main), type, Type),
	(
	    Type = tid('IO', 'IO')<@[tid('Prelude', void)<@[]] ->
	    make_name(qid(Module, main), Main),
	    portray_clause(( main :- tohnf(Main#R, R),
                                 tonf(R, _Rnf)))
	;
	    throw(error(compiler, main_has_to_be_IO_but_is(Type)))
	).

write_main_goal.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:
